@@ -9,14 +9,16 @@
 
 [@@@warning "-7-9-27-32-33"]
 
-open CamomileLibraryDefault.Camomile
 open Lwt_react
 open LTerm_dlist
 open LTerm_text
 open LTerm_geom
+open UTop
+open UTop_compat
 open UTop_token
 open UTop_styles
 open UTop_private
+open UTop.Private
 
 let return, (>>=) = Lwt.return, Lwt.(>>=)
 
@@ -39,7 +41,7 @@ let save_history () =
                         ?max_entries:!UTop.history_file_max_entries fn)
           (function
           | Unix.Unix_error (error, func, arg) ->
-              Lwt_log.error_f "cannot save history to %S: %s: %s" fn func (Unix.error_message error)
+              Logs_lwt.err (fun m -> m "cannot save history to %S: %s: %s" fn func (Unix.error_message error))
           | exn -> Lwt.fail exn)
 
 let init_history () =
@@ -54,8 +56,8 @@ let init_history () =
           (fun () -> LTerm_history.load UTop.history fn)
           (function
           | Unix.Unix_error (error, func, arg) ->
-              Lwt_log.error_f "cannot load history from %S: %s: %s"
-                              fn func (Unix.error_message error)
+              Logs_lwt.err (fun m -> m "cannot load history from %S: %s: %s"
+                                fn func (Unix.error_message error))
           | exn -> Lwt.fail exn)
 
 (* +-----------------------------------------------------------------+
@@ -77,23 +79,50 @@ let index_of_offset src ofs =
   in
   aux 0 0
 
-let convert_locs str locs = List.map (fun (a, b) -> (index_of_offset str a, index_of_offset str b)) locs
+let convert_loc str (a, b) = (index_of_offset str a, index_of_offset str b)
+
+let convert_locs str locs = List.map (fun (a, b) -> convert_loc str (a,b)) locs
+
+let get_line src line =
+  let rec aux line' ofs skipped =
+    if ofs >= String.length src then
+      ("", 0)
+    else if line' = line then
+      (String.sub src ofs (String.length src - ofs), skipped)
+    else
+      let ch, next_ofs = Zed_utf8.unsafe_extract_next src ofs in
+      if Zed_utf8.escaped_char ch = "\\n" then
+        aux (line' + 1) next_ofs (skipped + 1)
+      else
+        aux line' next_ofs (skipped + 1)
+  in
+  aux 1 0 0
+
+let convert_one_line str line ofs=
+  let selected, skipped = get_line str line in
+  index_of_offset selected ofs + skipped
+
+let convert_line str (start_ofs, end_ofs) lines =
+  (convert_one_line str lines.start start_ofs,
+  convert_one_line str lines.stop end_ofs)
+
+let convert_loc_line input locs lines =
+  List.map2 (fun loc line ->
+    match line with
+    | None ->
+      convert_loc input loc
+    | Some line ->
+      convert_line input loc line) locs lines
 
 (* +-----------------------------------------------------------------+
    | The read-line class                                             |
    +-----------------------------------------------------------------+ *)
 
-#if OCAML_VERSION >= (4, 04, 0)
-let ast_impl_kind = Pparse.Structure
-#else
-let ast_impl_kind = Config.ast_impl_magic_number
-#endif
-
 let preprocess input =
   match input with
     | Parsetree.Ptop_def pstr ->
         Parsetree.Ptop_def
-          (Pparse.apply_rewriters ~tool_name:"ocaml" ast_impl_kind pstr)
+          (Pparse.apply_rewriters ~tool_name:"ocaml" Pparse.Structure pstr)
     | _ -> input
 
 let parse_input_multi input =
@@ -113,7 +142,7 @@ let parse_input_multi input =
   in
   (result, Buffer.contents buf)
 
-let parse_and_check input eos_is_error =
+let parse_and_check input ~eos_is_error =
   let buf = Buffer.create 32 in
   let result =
     UTop.collect_formatters buf [Format.err_formatter]
@@ -127,8 +156,8 @@ let parse_and_check input eos_is_error =
                  match UTop.check_phrase phrase with
                    | None ->
                        UTop.Value phrase
-                   | Some (locs, msg) ->
-                       UTop.Error (convert_locs input locs, msg)
+                   | Some (locs, msg, lines) ->
+                       UTop.Error (convert_loc_line input locs lines, msg)
                with Pparse.Error error ->
                  Pparse.report_error Format.str_formatter error;
                  UTop.Error ([], "Error: " ^ Format.flush_str_formatter () ^ "\n"))
@@ -189,9 +218,8 @@ class read_phrase ~term = object(self)
         let input_utf8= Zed_string.to_utf8 input in
         (* Toploop does that: *)
         Location.reset ();
-        let eos_is_error = not !UTop.smart_accept in
         try
-          let result = parse_and_check input_utf8 eos_is_error in
+          let result = parse_and_check input_utf8 ~eos_is_error:false in
           return_value <- Some result;
           LTerm_history.add UTop.history input;
           let out, warnings = result in
@@ -208,7 +236,7 @@ class read_phrase ~term = object(self)
           return (LTerm_read_line.Result result)
         with UTop.Need_more ->
           (* Input not finished, continue. *)
-          self#insert (UChar.of_char '\n');
+          self#insert (Uchar.of_char '\n');
           self#exec ~keys actions
       end
     | actions ->
@@ -268,30 +296,100 @@ class read_phrase ~term = object(self)
 end
 
 (* +-----------------------------------------------------------------+
-   | Out phrase printing                                             |
+   | Handling of [@@toplevel_printer] attributes                     |
    +-----------------------------------------------------------------+ *)
 
-let fix_string str =
-  let len = String.length str in
-  let ofs, _, _ = Zed_utf8.next_error str 0 in
-  if ofs = len then
-    str
-  else begin
-    let buf = Buffer.create (len + 128) in
-    if ofs > 0 then Buffer.add_substring buf str 0 ofs;
-    let rec loop ofs =
-      Zed_utf8.add buf (UChar.of_char str.[ofs]);
-      let ofs1 = ofs + 1 in
-      let ofs2, _, _ = Zed_utf8.next_error str ofs1 in
-      if ofs1 < ofs2 then
-        Buffer.add_substring buf str ofs1 (ofs2 - ofs1);
-      if ofs2 < len then
-        loop ofs2
-      else
-        Buffer.contents buf
-    in
-    loop ofs
-  end
+module Autoprinter : sig
+  val scan_env : Format.formatter -> unit
+
+  val scan_cmis : Format.formatter -> unit
+end = struct
+  open Types
+
+  let cons_path path id =
+    let comp = Ident.name id in
+    match path with
+    | None -> Longident.Lident comp
+    | Some path -> Longident.Ldot (path, comp)
+
+  let is_auto_printer_attribute (attr : Parsetree.attribute) =
+    let name = attr.attr_name in
+    match name.txt with
+    | "toplevel_printer" | "ocaml.toplevel_printer" -> true
+    | _ -> false
+
+  let rec walk_sig pp ~path signature =
+    List.iter (walk_sig_item pp (Some path)) signature
+
+  and walk_sig_item pp path = function
+    | Sig_module (id, _, {md_type = mty; _}, _, _) ->
+        walk_mty pp (cons_path path id) mty
+    | Sig_value (id, vd, _) ->
+      if List.exists is_auto_printer_attribute vd.val_attributes then
+        Topdirs.dir_install_printer pp (cons_path path id)
+    | _ -> ()
+
+  and walk_mty pp path = function
+    | Mty_signature s -> walk_sig pp ~path s
+    | _ -> ()
+
+  let scan_cmis =
+    let new_cmis = ref [] in
+    UTop_compat.add_cmi_hook (fun cmi -> new_cmis := cmi :: !new_cmis );
+    fun pp ->
+      List.iter (fun (cmi : Cmi_format.cmi_infos) ->
+        walk_sig pp ~path:(Longident.Lident cmi.cmi_name) cmi.cmi_sign
+      ) !new_cmis;
+      new_cmis := []
+
+  let scan_env =
+    let last_globals = ref (Env.get_required_globals ()) in
+    let last_summary = ref Env.Env_empty in
+    fun pp ->
+      let env = !Toploop.toplevel_env in
+      let scan_module env id =
+        let name = Longident.Lident (Ident.name id) in
+        let path, {md_type; _} = Env.find_module_by_name name env in
+        if path = Path.Pident id then
+          walk_mty pp name md_type
+      in
+      let rec scan_globals last = function
+        | [] -> ()
+        | x when x == last -> ()
+        | x :: xs ->
+          scan_globals last xs;
+          scan_module env x
+      in
+      let rec scan_summary last = function
+        | Env.Env_empty -> ()
+        | x when x == last -> ()
+        | Env.Env_module (s, id, _, _) ->
+          scan_summary last s;
+          scan_module env id
+        | Env.Env_copy_types s
+        | Env.Env_value_unbound (s, _, _)
+        | Env.Env_module_unbound (s, _, _)
+        | Env.Env_persistent (s, _)
+        | Env.Env_value (s, _, _)
+        | Env.Env_type (s, _, _)
+        | Env.Env_extension (s, _, _)
+        | Env.Env_modtype (s, _, _)
+        | Env.Env_class (s, _, _)
+        | Env.Env_cltype (s, _, _)
+        | Env.Env_open (s, _)
+        | Env.Env_functor_arg (s, _)
+        | Env.Env_constraints (s, _) ->
+          scan_summary last s
+      in
+      let globals = Env.get_required_globals () in
+      let last_globals' = !last_globals in
+      last_globals := globals;
+      scan_globals last_globals' globals;
+      let summary = Env.summary env in
+      let last_summary' = !last_summary in
+      last_summary := summary;
+      scan_summary last_summary' summary
+end
 
 let render_out_phrase term string =
   if String.length string >= 100 * 1024 then
@@ -336,14 +434,9 @@ let map_items unwrap wrap items =
             (name, rs)
          | Outcometree.Osig_typext ({ Outcometree.oext_name = name}, _)
          | Outcometree.Osig_modtype (name, _)
-#if OCAML_VERSION < (4, 03, 0)
-         | Outcometree.Osig_value (name, _, _) ->
-            (name, Outcometree.Orec_not)
-#else
          | Outcometree.Osig_value { oval_name = name; _ } ->
             (name, Outcometree.Orec_not)
          | Outcometree.Osig_ellipsis -> ("", Outcometree.Orec_not)
-#endif
        in
        let keep =
          name = "" || name.[0] <> '_' ||
@@ -381,9 +474,7 @@ let map_items unwrap wrap items =
                  else
                    items
               | Outcometree.Osig_typext _
-#if OCAML_VERSION >= (4, 03, 0)
               | Outcometree.Osig_ellipsis
-#endif
               | Outcometree.Osig_modtype _
               | Outcometree.Osig_value _ ->
                  items
@@ -399,6 +490,7 @@ let print_out_signature pp items =
     orig_print_out_signature pp items
 
 let print_out_phrase pp phrase =
+  Autoprinter.scan_env pp;
   if UTop.get_hide_reserved () then
     let phrase =
       match phrase with
@@ -437,24 +529,9 @@ type rewrite_rule = {
   (* Whether the rule is enabled or not. *)
 }
 
-#if OCAML_VERSION < (4, 11, 0)
-let longident_parse= Longident.parse
-#else
-let longident_parse str=
-  let lexbuf= Lexing.from_string str in
-  Parse.longident lexbuf
-#endif
-
 let longident_lwt_main_run = Longident.Ldot (Longident.Lident "Lwt_main", "run")
 let longident_async_thread_safe_block_on_async_exn =
   Longident.(Ldot (Ldot (Lident "Async", "Thread_safe"), "block_on_async_exn"))
-let longident_unit = Longident.Lident "()"
-
-#if OCAML_VERSION >= (4, 03, 0)
-let nolabel = Asttypes.Nolabel
-#else
-let nolabel = ""
-#endif
 
 let rewrite_rules = [
   (* Rewrite Lwt.t expressions to Lwt_main.run <expr> *)
@@ -465,7 +542,7 @@ let rewrite_rules = [
     rewrite = (fun loc e ->
       let open Ast_helper in
       with_default_loc loc (fun () ->
-        Exp.apply (Exp.ident (with_loc loc longident_lwt_main_run)) [(nolabel, e)]
+        Exp.apply (Exp.ident (with_loc loc longident_lwt_main_run)) [(Nolabel, e)]
       )
     );
     enabled = UTop.auto_run_lwt;
@@ -483,24 +560,13 @@ let rewrite_rules = [
       with_default_loc loc (fun () ->
         Exp.apply
           (Exp.ident (with_loc loc longident_async_thread_safe_block_on_async_exn))
-          [(nolabel, Exp.fun_ nolabel None punit e)]
+          [(Nolabel, UTop_compat.Exp.fun_ ~loc punit e)]
       )
     );
     enabled = UTop.auto_run_async;
   }
 ]
 
-#if OCAML_VERSION >= (4, 10, 0)
-let lookup_type longident env =
-  Env.find_type_by_name longident env
-#elif OCAML_VERSION >= (4, 04, 0)
-let lookup_type longident env =
-  let path = Env.lookup_type longident env in
-  (path, Env.find_type path env)
-#else
-let lookup_type = Env.lookup_type
-#endif
-
 let rule_path rule =
   match rule.path_to_rewrite with
   | Some _ as x -> x
@@ -508,13 +574,13 @@ let rule_path rule =
     try
       let env = !Toploop.toplevel_env in
       let path =
-        match lookup_type rule.type_to_rewrite env with
-        | path, { Types.type_kind     = Types.Type_abstract
+        match Env.find_type_by_name rule.type_to_rewrite env with
+        | path, { Types.type_kind     = type_kind
                 ; Types.type_private  = Asttypes.Public
                 ; Types.type_manifest = Some ty
-                } -> begin
-            match Ctype.expand_head env ty with
-            | { Types.desc = Types.Tconstr (path, _, _) } -> path
+                } when type_kind = UTop_compat.abstract_type_kind -> begin
+            match get_desc (Ctype.expand_head env ty) with
+            | Types.Tconstr (path, _, _) -> path
             | _ -> path
           end
         | path, _ -> path
@@ -525,28 +591,11 @@ let rule_path rule =
     with _ ->
       None
 
-(* Returns whether the given path is persistent. *)
-let rec is_persistent_path = function
-  | Path.Pident id -> Ident.persistent id
-#if OCAML_VERSION >= (4, 08, 0)
-  | Path.Pdot (p, _) -> is_persistent_path p
-#else
-  | Path.Pdot (p, _, _) -> is_persistent_path p
-#endif
-  | Path.Papply (_, p) -> is_persistent_path p
-
 (* Check that the given long identifier is present in the environment
    and is persistent. *)
 let is_persistent_in_env longident =
-  let lookup_value=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_value_by_name
-#else
-    Env.lookup_value
-#endif
-  in
   try
-    is_persistent_path (fst (lookup_value longident !Toploop.toplevel_env))
+    is_persistent_path (fst (Env.find_value_by_name longident !Toploop.toplevel_env))
   with Not_found ->
     false
 
@@ -564,7 +613,7 @@ let is_eval = function
 
 (* Returns the rewrite rule associated to a type, if any. *)
 let rule_of_type typ =
-  match (Ctype.expand_head !Toploop.toplevel_env typ).Types.desc with
+  match get_desc (Ctype.expand_head !Toploop.toplevel_env typ) with
   | Types.Tconstr (path, _, _) -> begin
       try
         Some (List.find (fun rule -> rule_matches rule path) rewrite_rules)
@@ -590,12 +639,12 @@ let rewrite_str_item pstr_item tstr_item =
       pstr_item
 
 let type_structure env pstr =
-#if OCAML_VERSION >= (4, 12, 0)
+#if OCAML_VERSION >= (4, 14, 0)
+  let tstr, _, _, _, _ = Typemod.type_structure env pstr in
+#elif OCAML_VERSION >= (4, 12, 0)
   let tstr, _, _, _ = Typemod.type_structure env pstr in
-#elif OCAML_VERSION >= (4, 08, 0)
-  let tstr, _, _, _ = Typemod.type_structure env pstr Location.none in
 #else
-  let tstr, _, _ = Typemod.type_structure env pstr Location.none in
+  let tstr, _, _, _ = Typemod.type_structure env pstr Location.none in
 #endif
   tstr
 
@@ -613,24 +662,12 @@ let rewrite phrase =
 let add_let binding_name def =
   let open Parsetree in
   match def with
-  | { pstr_desc = Pstr_eval (expr, attr); pstr_loc } ->
-    {
-      pstr_loc;
-      pstr_desc = Pstr_value (Asttypes.Nonrecursive, [
-        {
-          pvb_pat = {
-            ppat_desc = Ppat_var { txt = binding_name; loc = pstr_loc; };
-#if OCAML_VERSION >= (4, 08, 0)
-            ppat_loc_stack= [];
-#endif
-            ppat_loc = pstr_loc;
-            ppat_attributes = [];
-          };
-          pvb_expr = expr;
-          pvb_attributes = attr;
-          pvb_loc = pstr_loc;
-        }]);
-    }
+  | { pstr_desc = Pstr_eval (expr, attrs); pstr_loc = loc } ->
+      Ast_helper.Str.value ~loc Nonrecursive
+      [
+        let pat = Ast_helper.Pat.var ~loc { txt = binding_name; loc } in
+        Ast_helper.Vb.mk ~loc ~attrs pat expr
+      ]
   | _ ->
     def
 
@@ -641,84 +678,11 @@ let bind_expressions name phrase =
     | Parsetree.Ptop_dir _ ->
       phrase
 
-(* +-----------------------------------------------------------------+
-   | Handling of [@@toplevel_printer] attributes                     |
-   +-----------------------------------------------------------------+ *)
-
-#if OCAML_VERSION >= (4, 04, 0)
-
-#if OCAML_VERSION >= (4, 09, 0)
-  module Persistent_signature = Persistent_env.Persistent_signature
-#else
-  module Persistent_signature = Env.Persistent_signature
-#endif
-
-let execute_phrase =
-  let new_cmis = ref []in
-  let default_load = !Persistent_signature.load in
-  let load ~unit_name =
-    let res = default_load ~unit_name in
-    (match res with None -> () | Some x -> new_cmis := x.cmi :: !new_cmis);
-    res
-  in
-  Persistent_signature.load := load;
-
-  let rec collect_printers path signature acc =
-    List.fold_left (fun acc item ->
-      match (item : Types.signature_item) with
-#if OCAML_VERSION >= (4, 08, 0)
-      | Sig_module (id, _, {md_type = Mty_signature s; _}, _, _) ->
-#else
-      | Sig_module (id, {md_type = Mty_signature s; _}, _) ->
-#endif
-        collect_printers (Longident.Ldot (path, Ident.name id)) s acc
-#if OCAML_VERSION >= (4, 08, 0)
-      | Sig_value (id, vd, _) ->
-#else
-      | Sig_value (id, vd) ->
-#endif
-#if OCAML_VERSION >= (4, 08, 0)
-        if List.exists (fun attr->
-          let open Parsetree in
-          match attr.attr_name with
-          | {Asttypes.txt = "toplevel_printer" | "ocaml.toplevel_printer"; _} ->
-#else
-        if List.exists (function
-          | {Asttypes.txt = "toplevel_printer" | "ocaml.toplevel_printer"; _},
-            _ ->
-#endif
-            true
-          | _ -> false)
-          vd.val_attributes
-        then
-          Longident.Ldot (path, Ident.name id) :: acc
-        else acc
-      | _ -> acc)
-      acc signature
-  in
-
-  let acknowledge_new_cmis () =
-    let l = !new_cmis in
-    new_cmis := [];
-    let printers =
-    List.fold_left (fun acc (cmi : Cmi_format.cmi_infos) ->
-        collect_printers (Longident.Lident cmi.cmi_name) cmi.cmi_sign acc )
-      [] l
-    in
-    List.iter (Topdirs.dir_install_printer Format.err_formatter) printers
-  in
-
-  fun b pp phrase ->
-    acknowledge_new_cmis ();
-    let res = Toploop.execute_phrase b pp phrase in
-    acknowledge_new_cmis ();
-    res
-
-#else
-
-let execute_phrase = Toploop.execute_phrase
-
-#endif
+let execute_phrase b ppf phrase =
+  Autoprinter.scan_cmis ppf;
+  let res = Toploop.execute_phrase b ppf phrase in
+  Autoprinter.scan_cmis ppf;
+  res
 
 (* +-----------------------------------------------------------------+
    | Main loop                                                       |
@@ -1025,7 +989,7 @@ module Emacs(M : sig end) = struct
   let process_input add_to_history eos_is_error =
     let input = read_data () in
     let input_zed= Zed_string.unsafe_of_utf8 input in
-    let result, warnings = parse_and_check input eos_is_error in
+    let result, warnings = parse_and_check input ~eos_is_error in
     match result with
       | UTop.Value phrase ->
           send "accept" "";
@@ -1051,7 +1015,7 @@ module Emacs(M : sig end) = struct
     let typecheck phrase =
       match UTop.check_phrase phrase with
         | None -> None
-        | Some (locs, msg) -> Some (convert_locs input locs, msg)  (* FIXME *)
+        | Some (locs, msg, lines) -> Some (convert_loc_line input locs lines, msg)
     in
     match result with
       | UTop.Value phrases ->
@@ -1201,25 +1165,16 @@ end
    | Extra macros                                                    |
    +-----------------------------------------------------------------+ *)
 
-let typeof sid =
-  let id  = longident_parse sid in
-  let env = !Toploop.toplevel_env in
-#if OCAML_VERSION >= (4, 10, 0)
-  let lookup_value= Env.find_value_by_name
-  and lookup_label= Env.find_label_by_name
-  and lookup_modtype= Env.find_modtype_by_name
-  and lookup_module id env =
-    let path, decl = Env.find_module_by_name id env in
-    (path, decl.md_type)
+module Printtyp =
+#if OCAML_VERSION >= (5, 3, 0)
+  Out_type
 #else
-  let lookup_value= Env.lookup_value
-  and lookup_label= Env.lookup_label
-  and lookup_modtype= Env.lookup_modtype
-  and lookup_module id env =
-    let path = Env.lookup_module id env ~load:true in
-    (path, (Env.find_module path env).md_type)
+  Printtyp
 #endif
-  in
+
+let typeof sid =
+  let id = Parse.longident (Lexing.from_string sid) in
+  let env = !Toploop.toplevel_env in
   let from_type_desc = function
     | Types.Tconstr (path, _, _) ->
       let typ_decl = Env.find_type path env in
@@ -1228,65 +1183,37 @@ let typeof sid =
   in
   let out_sig_item =
     try
-      let (path, ty_decl) = lookup_type id env in
-#if OCAML_VERSION >= (4, 08, 0)
+      let (path, ty_decl) = Env.find_type_by_name id env in
       let id = Ident.create_local (Path.name path) in
-#else
-      let id = Ident.create (Path.name path) in
-#endif
       Some (Printtyp.tree_of_type_declaration id ty_decl Types.Trec_not)
     with Not_found ->
     try
-      let (path, val_descr) = lookup_value id env in
-#if OCAML_VERSION >= (4, 08, 0)
+      let (path, val_descr) = Env.find_value_by_name id env in
       let id = Ident.create_local (Path.name path) in
-#else
-      let id = Ident.create (Path.name path) in
-#endif
       Some (Printtyp.tree_of_value_description id val_descr)
     with Not_found ->
     try
-      let lbl_desc = lookup_label id env in
-      let (path, ty_decl) = from_type_desc lbl_desc.Types.lbl_res.Types.desc in
-#if OCAML_VERSION >= (4, 08, 0)
+      let lbl_desc = Env.find_label_by_name id env in
+      let (path, ty_decl) = from_type_desc (get_desc lbl_desc.Types.lbl_res) in
       let id = Ident.create_local (Path.name path) in
-#else
-      let id = Ident.create (Path.name path) in
-#endif
       Some (Printtyp.tree_of_type_declaration id ty_decl Types.Trec_not)
     with Not_found ->
     try
-      let path, mod_typ = lookup_module id env in
-#if OCAML_VERSION >= (4, 08, 0)
+      let path, {Types.md_type; _} = Env.find_module_by_name id env in
       let id = Ident.create_local (Path.name path) in
-#else
-      let id = Ident.create (Path.name path) in
-#endif
-      Some (Printtyp.tree_of_module id mod_typ Types.Trec_not)
+      Some (Printtyp.tree_of_module id md_type Types.Trec_not)
     with Not_found ->
     try
-      let (path, mty_decl) = lookup_modtype id env in
-#if OCAML_VERSION >= (4, 08, 0)
+      let (path, mty_decl) = Env.find_modtype_by_name id env in
       let id = Ident.create_local (Path.name path) in
-#else
-      let id = Ident.create (Path.name path) in
-#endif
       Some (Printtyp.tree_of_modtype_declaration id mty_decl)
     with Not_found ->
     try
-      #if OCAML_VERSION >= (4, 10, 0)
       let cstr_desc = Env.find_constructor_by_name id env in
-      #else
-      let cstr_desc = Env.lookup_constructor id env in
-      #endif
       match cstr_desc.Types.cstr_tag with
       | _ ->
-        let (path, ty_decl) = from_type_desc cstr_desc.Types.cstr_res.Types.desc in
-#if OCAML_VERSION >= (4, 08, 0)
+        let (path, ty_decl) = from_type_desc (get_desc cstr_desc.Types.cstr_res) in
         let id = Ident.create_local (Path.name path) in
-#else
-        let id = Ident.create (Path.name path) in
-#endif
         Some (Printtyp.tree_of_type_declaration id ty_decl Types.Trec_not)
     with Not_found ->
       None
@@ -1298,14 +1225,24 @@ let typeof sid =
   | Some osig ->
     let buf = Buffer.create 128 in
     let pp = Format.formatter_of_buffer buf in
+#if OCAML_VERSION >= (5, 3, 0)
+    Format.fprintf pp "%a" (Format_doc.compat !Toploop.print_out_signature) [osig];
+#else
     !Toploop.print_out_signature pp [osig];
+#endif
     Format.pp_print_newline pp ();
     let str = Buffer.contents buf in
     Lwt_main.run (Lazy.force LTerm.stdout >>= fun term -> render_out_phrase term str)
 
+let default_info = {
+  Toploop.section = "UTop";
+  doc = ""; (* TODO: have some kind of documentation *)
+}
+
 let () =
-  Hashtbl.add Toploop.directive_table "typeof"
+  Toploop.add_directive "typeof"
     (Toploop.Directive_string typeof)
+    default_info
 
 (* +-----------------------------------------------------------------+
    | Entry point                                                     |
@@ -1315,13 +1252,13 @@ let emacs_mode = ref false
 let preload = ref []
 
 let prepare () =
-  Toploop.set_paths ();
+  toploop_set_paths ();
   try
     let ok =
       List.for_all
         (function
           | `Packages l -> UTop.require l; true
-          | `Object fn -> Topdirs.load_file Format.err_formatter fn)
+          | `Object fn -> toploop_load_file Format.err_formatter fn)
         (List.rev !preload)
     in
     if ok then !Toploop.toplevel_startup_hook ();
@@ -1334,7 +1271,6 @@ let prepare () =
       Format.eprintf "Uncaught exception: %s\n" (Printexc.to_string exn);
       false
 
-#if OCAML_VERSION >= (4, 09, 0)
 external caml_sys_modify_argv : string array -> unit =
   "caml_sys_modify_argv"
 let override_argv () =
@@ -1342,14 +1278,6 @@ let override_argv () =
   let copy = Array.init len (fun i -> Sys.argv.(i+ !Arg.current)) in
   caml_sys_modify_argv copy;
   Arg.current := 0
-#else
-let override_argv () =
-  let len = Array.length Sys.argv - !Arg.current in
-  Array.blit Sys.argv !Arg.current Sys.argv 0 len;
-  Obj.truncate (Obj.repr Sys.argv) len;
-  Arg.current := 0
-#endif
-
 
 let run_script name =
   (* To prevent message from camlp4 *)
@@ -1358,7 +1286,7 @@ let run_script name =
   override_argv ();
   Toploop.initialize_toplevel_env ();
   Location.input_name := UTop.input_name;
-  if Toploop.use_silently Format.err_formatter name then
+  if toploop_use_silently Format.err_formatter name then
     exit 0
   else
     exit 2
@@ -1381,11 +1309,7 @@ let print_version_num () =
 let autoload = ref true
 
 let args = Arg.align [
-#if OCAML_VERSION >= (4, 08, 0)
   "-absname", Arg.Set Clflags.absname, " Show absolute filenames in error message";
-#else
-  "-absname", Arg.Set Location.absname, " Show absolute filenames in error message";
-#endif
   "-I", Arg.String (fun dir ->  Clflags.include_dirs := dir :: !Clflags.include_dirs), "<dir> Add <dir> to the list of include directories";
   "-init", Arg.String (fun s -> Clflags.init_file := Some s), "<file> Load <file> instead of default init file";
   "-labels", Arg.Clear Clflags.classic, " Use commuting label mode";
@@ -1395,20 +1319,18 @@ let args = Arg.align [
   "-nostdlib", Arg.Set Clflags.no_std_include, " Do not add default directory to the list of include directories";
   "-ppx", Arg.String (fun ppx -> Clflags.all_ppx := ppx :: !Clflags.all_ppx), "<command> Pipe abstract syntax trees through preprocessor <command>";
   "-principal", Arg.Set Clflags.principal, " Check principality of type inference";
+#if OCAML_VERSION < (5, 0, 0)
   "-safe-string", Arg.Clear Clflags.unsafe_string, " Make strings immutable";
+#endif
   "-short-paths", Arg.Clear Clflags.real_paths, " Shorten paths in types (the default)";
   "-no-short-paths", Arg.Set Clflags.real_paths, " Do not shorten paths in types";
   "-rectypes", Arg.Set Clflags.recursive_types, " Allow arbitrary recursive types";
   "-stdin", Arg.Unit (fun () -> run_script ""), " Read script from standard input";
   "-strict-sequence", Arg.Set Clflags.strict_sequence, " Left-hand part of a sequence must have type unit";
-#if OCAML_VERSION >= (4, 08, 0)
   "-unsafe", Arg.Set Clflags.unsafe, " Do not compile bounds checking on array and string access";
-#else
-  "-unsafe", Arg.Set Clflags.fast, " Do not compile bounds checking on array and string access";
-#endif
   "-version", Arg.Unit print_version, " Print version and exit";
   "-vnum", Arg.Unit print_version_num, " Print version number and exit";
-  "-w", Arg.String (Warnings.parse_options false),
+  "-w", Arg.String (fun opt -> ignore (Warnings.parse_options false opt)),
   Printf.sprintf
     "<list>  Enable or disable warnings according to <list>:\n\
     \        +<spec>   enable warnings in <spec>\n\
@@ -1419,7 +1341,7 @@ let args = Arg.align [
     \        <num1>..<num2>    a range of consecutive warning numbers\n\
     \        <letter>          a predefined set\n\
     \     default setting is %S" Warnings.defaults_w;
-  "-warn-error", Arg.String (Warnings.parse_options true),
+  "-warn-error", Arg.String (fun opt -> ignore (Warnings.parse_options true opt)),
   Printf.sprintf
     "<list>  Enable or disable error status for warnings according to <list>\n\
     \     See option -w for the syntax of <list>.\n\
@@ -1453,7 +1375,7 @@ let load_init_files dir =
   Array.iter
     (fun fn ->
        if Filename.check_suffix fn ".ml" then
-         ignore (Toploop.use_silently Format.err_formatter (Filename.concat dir fn) : bool))
+         ignore (toploop_use_silently Format.err_formatter (Filename.concat dir fn) : bool))
     files
 ;;
 
@@ -1485,7 +1407,7 @@ let common_init ~initial_env =
            None
          )
      | None ->
-         if Sys.file_exists ".ocamlinit" then
+         if Sys.file_exists ".ocamlinit" && Sys.getcwd () <> LTerm_resources.home then
            Some ".ocamlinit"
          else
            let xdg_fn = LTerm_resources.xdgbd_file ~loc:LTerm_resources.Config "utop/init.ml" in
@@ -1501,7 +1423,7 @@ let common_init ~initial_env =
   (match init_fn with
    | None -> ()
    | Some fn ->
-     ignore (Toploop.use_silently Format.err_formatter fn : bool));
+     ignore (toploop_use_silently Format.err_formatter fn : bool));
   (* Load history after the initialization file so the user can change
      the history file name. *)
   Lwt_main.run (init_history ());
@@ -1524,15 +1446,18 @@ let load_inputrc () =
     LTerm_inputrc.load
     (function
     | Unix.Unix_error (error, func, arg) ->
-      Lwt_log.error_f "cannot load key bindings from %S: %s: %s" LTerm_inputrc.default func (Unix.error_message error)
+      Logs_lwt.err (fun m -> m "cannot load key bindings from %S: %s: %s" LTerm_inputrc.default func (Unix.error_message error))
     | LTerm_inputrc.Parse_error (fname, line, msg) ->
-      Lwt_log.error_f "error in key bindings file %S, line %d: %s" fname line msg
+      Logs_lwt.err (fun m -> m "error in key bindings file %S, line %d: %s" fname line msg)
     | exn -> Lwt.fail exn)
 
 let protocol_version = 1
 
 let main_aux ~initial_env =
   Arg.parse args file_argument usage;
+#if OCAML_VERSION >= (5, 0, 0) && OCAML_VERSION < (5, 1, 0)
+  Topcommon.load_topdirs_signature ();
+#endif
   if not (prepare ()) then exit 2;
   if !emacs_mode then begin
     Printf.printf "protocol-version:%d\n%!" protocol_version;
@@ -1576,10 +1501,13 @@ let main_internal ~initial_env =
     main_aux ~initial_env
   with exn ->
     (match exn with
+#if OCAML_VERSION >= (4,12,0)
+       | Compenv.Exit_with_status e -> exit e
+#endif
        | Unix.Unix_error (error, func, "") ->
            Printf.eprintf "%s: %s: %s\n" app_name func (Unix.error_message error)
        | Unix.Unix_error (error, func, arg) ->
-           Printf.eprintf "%s: %s(%S): %s\n" app_name func arg (Unix.error_message error)
+         Printf.eprintf "%s: %s(%S): %s\n" app_name func arg (Unix.error_message error)
        | exn ->
            Printf.eprintf "Fatal error: exception %s\n" (Printexc.to_string exn));
     Printexc.print_backtrace stderr;
@@ -1592,18 +1520,10 @@ type value = V : string * _ -> value
 
 exception Found of Env.t
 
-#if OCAML_VERSION >= (4, 03, 0)
 let get_required_label name args =
   match List.find (fun (lab, _) -> lab = Asttypes.Labelled name) args with
   | _, x -> x
   | exception Not_found -> None
-#else
-let get_required_label name args =
-  match List.find (fun (lab, _, k) -> lab = "loc" && k = Typedtree.Required) args with
-  | _, x, _ -> x
-  | _ -> None
-  | exception Not_found -> None
-#endif
 
 let walk dir ~init ~f =
   let rec loop dir acc =
@@ -1618,12 +1538,19 @@ let walk dir ~init ~f =
   | exception Unix.Unix_error(ENOENT, _, _) -> init
   | _ -> loop dir init
 
+let iter_structure expr =
+  let next iterator e = Tast_iterator.default_iterator.expr iterator e in
+  let expr iterator = expr (next iterator) in
+  let iter = { Tast_iterator.default_iterator with expr } in
+  iter.structure iter
+
 let interact ?(search_path=[]) ?(build_dir="_build") ~unit ~loc:(fname, lnum, cnum, _)
       ~values =
   let search_path = walk build_dir ~init:search_path ~f:(fun dir acc -> dir :: acc) in
   let cmt_fname =
     try
-      Misc.find_in_path_uncap search_path (unit ^ ".cmt")
+      UTop_compat.find_in_path_normalized
+        search_path (unit ^ ".cmt")
     with Not_found ->
       Printf.ksprintf failwith "%s.cmt not found in search path!" unit
   in
@@ -1646,20 +1573,7 @@ let interact ?(search_path=[]) ?(build_dir="_build") ~unit ~loc:(fname, lnum, cn
           end
         | _ -> next e
   in
-#if OCAML_VERSION >= (4,09,0)
-  let next iterator e = Tast_iterator.default_iterator.expr iterator e in
-  let expr iterator = expr (next iterator) in
-  let iter = { Tast_iterator.default_iterator with expr } in
-  let search = iter.structure iter in
-#else
-  let module Search =
-    TypedtreeIter.MakeIterator(struct
-      include TypedtreeIter.DefaultIteratorArgument
-
-      let enter_expression = expr ignore
-     end) in
-    let search = Search.iter_structure in
-#endif
+  let search = iter_structure expr in
   try
     begin match cmt_infos.cmt_annots with
     | Implementation st -> search st
@@ -1668,7 +1582,8 @@ let interact ?(search_path=[]) ?(build_dir="_build") ~unit ~loc:(fname, lnum, cn
     failwith "Couldn't find location in cmt file"
   with Found env ->
   try
-    List.iter Topdirs.dir_directory (search_path @ cmt_infos.cmt_loadpath);
+    let visible_paths = UTop_compat.visible_paths_for_cmt_infos cmt_infos in
+    List.iter Topdirs.dir_directory (search_path @ visible_paths);
     let env = Envaux.env_of_only_summary env in
     List.iter (fun (V (name, v)) -> Toploop.setvalue name (Obj.repr v)) values;
     main_internal ~initial_env:(Some env)
@@ -1680,6 +1595,10 @@ let () =
   Location.register_error_of_exn
     (function
       | Envaux.Error err ->
+#if OCAML_VERSION >= (5,3,0)
+        Some (Location.error_of_printer_file Envaux.report_error_doc err)
+#else
         Some (Location.error_of_printer_file Envaux.report_error err)
+#endif
       | _ -> None
     )
